import numpy as np
import torch
from sklearn.model_selection import train_test_split


def get_X(N=1_000, I=50, Q=30, N_test=200):
    N = N + N_test

    X = torch.randn([N * I, Q], dtype=torch.float32)

    X = (X - X.mean(0)) / X.std(0)
    X = X / np.sqrt(X.shape[1])
    X = X.reshape([N, I, Q])

    return X


def simulate(X, v_beta=0.5, v_gamma=0.8, b=-1, F=None, P=1):
    if F is None:
        F = torch.ones([X.shape[0], 1])

    # simulate single phenotype
    K = F.shape[1]
    b = b * torch.ones([K, P])
    v_beta = v_beta * torch.ones(P)
    v_gamma = v_gamma * torch.ones(P)

    # sample weights
    gamma = torch.randn((X.shape[2], v_gamma.shape[0]))
    _w = torch.einsum("nik,kp->nip", X, gamma)
    _scale_w = torch.sqrt(v_gamma / _w.var([0, 1]))
    gamma = _scale_w[None, :] * gamma
    _w = _scale_w[None, None, :] * _w

    w = torch.softmax(_w, dim=1)

    # sample z
    beta = torch.randn((X.shape[2], v_beta.shape[0]))
    beta = beta / torch.sqrt((beta**2).mean(0, keepdim=True))
    z = torch.einsum("nik,kp->nip", X, beta)
    u = torch.einsum("nip,nip->np", w, z)
    u = (u - u.mean(0)) / u.std(0)
    u = torch.sqrt(v_beta) * u
    beta = torch.sqrt(v_beta) * beta

    # compute rates
    logits = F.mm(b) + u

    # sample Y
    Y = torch.distributions.Binomial(2, logits=logits).sample()

    return F, Y, u, w


def split_data(Xs, test_size=200, val_size=0.0, test_rs=127, val_rs=412):
    idxs_all = np.arange(Xs[0].shape[0])
    idxs = {}
    idxs["train_val"], idxs["test"] = train_test_split(idxs_all, test_size=test_size, random_state=test_rs)

    if not np.isclose(val_size, 0):
        idxs["train"], idxs["val"] = train_test_split(idxs["train_val"], test_size=val_size, random_state=val_rs)
    else:
        idxs["train"] = idxs["train_val"]
    del idxs["train_val"]

    outs = []
    for X in Xs:
        out = {}
        for key in idxs.keys():
            out[key] = X[idxs[key]]
        outs.append(out)
    return outs


def load_simulation(seed=42, N=1_000, I=50, Q=30, N_test=200, P=1, v_beta=0.5, v_gamma=0.8, b=-1, F=None):
    np.random.seed(seed)
    torch.manual_seed(seed)

    X = get_X(N, I, Q, N_test)
    data = simulate(X, v_beta, v_gamma, b, F, P)
    X, F, Y, u, w = split_data((X, *data))
    return X, F, Y, u, w



import numpy as np
import scipy.linalg as la
import statsmodels.api as sm
import torch
from sklearn.linear_model import LogisticRegressionCV
from tqdm.auto import trange

from mixmil.data import xgower_factor


def regressOut(Y, X, return_b=False, return_pinv=False):
    """
    regresses out X from Y
    """
    Xd = la.pinv(X)
    b = Xd.dot(Y)
    Y_out = Y - X.dot(b)
    out = [Y_out]
    if return_b:
        out.append(b)
    if return_pinv:
        out.append(Xd)
    return out if len(out) > 1 else out[0]


def _get_single_binomial_init_params(X, F, y):
    ident = np.zeros((X.shape[1]), dtype=int)
    model = sm.BinomialBayesMixedGLM(y, F, X, ident).fit_vb()

    u = np.dot(X, model.vc_mean)[::2]

    _scale = u.std() / np.sqrt((model.vc_mean**2).mean(0))
    mu_beta = _scale * model.vc_mean
    sd_beta = _scale * model.vc_sd
    var_z = (mu_beta**2 + sd_beta**2).mean().reshape(1)
    alpha = model.fe_mean

    return mu_beta, sd_beta, var_z, alpha


def get_binomial_init_params(X, F, Y):
    results = [_get_single_binomial_init_params(X, F, Y[:, p]) for p in trange(Y.shape[1], desc="GLMM Init")]

    mu_beta, sd_beta, var_z, alpha = [_list2tensor(listo) for listo in zip(*results)]

    return mu_beta, sd_beta, var_z, alpha


def get_lr_init_params(X, Y, b, Fiv):
    model = LogisticRegressionCV(
        Cs=10,
        fit_intercept=True,
        penalty="l2",
        multi_class="multinomial",
        solver="lbfgs",
        n_jobs=8,
        verbose=0,
        random_state=42,
        max_iter=1000,
        refit=True,
    )

    model.fit(X, Y.ravel())

    alpha = model.intercept_[None]
    beta = model.coef_.T

    # Compute bag prediction u and reparametrize
    u = X.dot(beta)
    um = u.mean(0)[None]
    us = u.std(0)[None]
    alpha = alpha + um
    mu_beta = us * beta / np.sqrt((beta**2).mean(0)[None])
    sd_beta = np.sqrt(0.1 * (mu_beta**2).mean()) * np.ones_like(mu_beta)

    alpha = Fiv.dot(np.ones((Fiv.shape[1], 1))).dot(alpha) - b.dot(mu_beta)

    # init prior
    var_z = (mu_beta**2 + sd_beta**2).mean().reshape(1, 1)

    return [torch.Tensor(el) for el in (mu_beta, sd_beta, var_z, alpha)]


def _list2tensor(_list):
    return torch.Tensor(np.stack(_list, axis=1))


def get_init_params(Xs, F, Y, likelihood, n_trials):
    Xm = np.concatenate([x.mean(0, keepdims=True) for x in Xs], axis=0)
    Fe, Ye = F.numpy(), Y.long().numpy()

    if likelihood == "binomial":
        Xm = (Xm - Xm.mean(0, keepdims=True)) / xgower_factor(Xm)

        if n_trials == 2:
            Xm, Fe = Xm.repeat(2, axis=0), Fe.repeat(2, axis=0)
            to_expanded = np.array(([[0, 0], [1, 0], [1, 1]]))
            Ye = to_expanded[Y.long().numpy().T].transpose(1, 2, 0).reshape(-1, Y.shape[1])

        mu_z, sd_z, var_z, alpha = get_binomial_init_params(Xm, Fe, Ye)

    elif likelihood == "categorical":
        Xm, b, Fiv = regressOut(Xm, Fe, return_b=True, return_pinv=True)
        Xm = (Xm - Xm.mean(0, keepdims=True)) / (Xm.std(0, keepdims=True) * np.sqrt(Xm.shape[-1]))

        mu_z, sd_z, var_z, alpha = get_lr_init_params(Xm, Ye, b, Fiv)

    return mu_z, sd_z, var_z, alpha



"""
Script adapted from
https://github.com/binli123/dsmil-wsi/blob/master/download.py

Script and data provided by the DSMIL authors:

@inproceedings{li2021dual,
  title={Dual-stream multiple instance learning network for whole slide image classification with self-supervised contrastive learning},
  author={Li, Bin and Li, Yin and Eliceiri, Kevin W},
  booktitle={Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition},
  pages={14318--14328},
  year={2021}
}
"""

import argparse
import os
import urllib.request
import zipfile

from tqdm import tqdm


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def unzip_data(zip_path, data_path):
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(data_path)


DATASET_URLS = {
    "camelyon16": "https://uwmadison.box.com/shared/static/l9ou15iwup73ivdjq0bc61wcg5ae8dwe.zip",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", type=str, default="camelyon16", help="Dataset to be downloaded: camelyon16")
    parser.add_argument("--keep-zip", action="store_true", help="Keep the downloaded zip file")
    args = parser.parse_args()

    assert args.dataset in DATASET_URLS, f"Dataset {args.dataset} not found"

    print(f"downloading dataset: {args.dataset}")
    unzip_dir = f"data/{args.dataset}"
    zip_file_path = f"data/{args.dataset}-dataset.zip"
    os.makedirs(unzip_dir, exist_ok=True)
    download_url(DATASET_URLS[args.dataset], zip_file_path)
    unzip_data(zip_file_path, unzip_dir)

    if not args.keep_zip:
        os.remove(f"{args.dataset}-dataset.zip")

    print("done!")


if __name__ == "__main__":
    main()



import numpy as np
import torch
from torch.distributions import LowRankMultivariateNormal


def get_params(vc_mean, vc_sd, n_outs, n_vars, mean_field):
    mu_z = torch.Tensor(vc_mean)
    mu_u = torch.zeros_like(mu_z)
    mu = torch.cat([mu_u, mu_z], 1)
    sd_z = torch.Tensor(vc_sd)
    sd_u = torch.sqrt(0.1 * torch.ones_like(sd_z))

    if mean_field:
        cov_factor = torch.zeros(n_outs, n_vars, 1)
        cov_logdiag = 2.0 * torch.log(torch.cat([sd_u, sd_z], 1))
    else:
        diag = torch.diag_embed(torch.cat([sd_u, sd_z], 1))
        cov_factor = 1e-4 * torch.randn(diag.shape) + diag
        cov_logdiag = np.log(1e-4) * torch.ones(n_outs, n_vars)

    return mu, cov_factor, cov_logdiag


class GaussianVariationalPosterior(torch.nn.Module):
    def __init__(self, n_vars, n_outs, mean_field=True, init_params=None):
        super().__init__()
        self.n_vars = n_vars
        self.n_outs = n_outs
        self.mean_field = mean_field

        if init_params is not None:
            mu_z, sd_z, *_ = init_params
            mu_z = mu_z.T
            sd_z = sd_z.T
        else:
            mu_z = 1e-3 * torch.randn(n_outs, n_vars // 2)
            sd_z = 1e-3 * torch.randn(n_outs, n_vars // 2)

        mu, cov_factor, cov_logdiag = get_params(mu_z, sd_z, n_outs, n_vars, mean_field)

        self.mu = torch.nn.Parameter(mu)
        if mean_field:
            self.register_buffer("cov_factor", cov_factor)
            self.cov_logdiag = torch.nn.Parameter(cov_logdiag)
        else:
            self.cov_factor = torch.nn.Parameter(cov_factor)
            self.register_buffer("cov_logdiag", cov_logdiag)

    @property
    def distribution(self):
        return LowRankMultivariateNormal(self.mu, self.cov_factor, torch.exp(self.cov_logdiag))

    @property
    def q_mu(self):
        return self.mu.T

    def sample(self, n_samples):
        return self.distribution.rsample([n_samples]).permute([2, 1, 0])

    def extra_repr(self) -> str:
        return f"n_vars=2*{self.n_vars//2}, n_outs={self.n_outs}, mean_field={self.mean_field}"



import numpy as np
import pytest
import torch

from mixmil import MixMIL


@pytest.fixture
def mock_data_binomial():
    Xs = [torch.randn(10, 3) for _ in range(5)]  # List of tensors
    F = torch.randn(5, 4)  # Fixed effects
    Y = torch.randint(0, 2, (5, 1))  # Labels for binomial
    return Xs, F, Y


@pytest.fixture
def mock_data_categorical():
    N, Q, K = 50, 10, 4
    bag_sizes = torch.randint(5, 15, (N,))
    Xs = [torch.randn(bag_sizes[n], Q) for n in range(N)]  # List of tensors
    F = torch.randn(N, K)  # Fixed effects
    Y = torch.randint(0, 5, (N, 1))  # Labels for categorical
    return Xs, F, Y


def test_init_with_mean_model_binomial(mock_data_binomial):
    Xs, F, Y = mock_data_binomial
    model = MixMIL.init_with_mean_model(Xs, F, Y, likelihood="binomial", n_trials=2)
    assert isinstance(model, MixMIL)
    assert model.likelihood_name == "binomial"
    assert model.n_trials == 2
    assert model.log_sigma_u.numel() == 1


def test_init_with_mean_model_categorical(mock_data_categorical):
    Xs, F, Y = mock_data_categorical
    model = MixMIL.init_with_mean_model(Xs, F, Y, likelihood="categorical")
    assert isinstance(model, MixMIL)
    assert model.likelihood_name == "categorical"
    assert model.n_trials is None
    assert model.log_sigma_u.numel() == len(np.unique(Y))  # separate prior for each class


def test_initialization():
    model = MixMIL(Q=10, K=5, P=2, likelihood="binomial", n_trials=2)
    assert model.Q == 10
    assert model.alpha.shape == (5, 2)


@pytest.mark.parametrize(
    "Q, K, P, likelihood, n_trials, mean_field",
    [
        (10, 5, 2, "categorical", None, True),
        (10, 5, 2, "categorical", None, False),
        (10, 5, 2, "binomial", 1, True),
        (10, 5, 2, "binomial", 2, False),
        (10, 5, 1, "binomial", 2, False),
    ],
)
def test_init_model(Q, K, P, likelihood, n_trials, mean_field):
    MixMIL(Q=Q, K=K, P=P, likelihood=likelihood, n_trials=n_trials, mean_field=mean_field)



import numpy as np
import torch
from torch.utils.data import Dataset

from mixmil.simulation import load_simulation


def setup_scatter(Xs):
    device = Xs[0].device
    x = torch.cat(Xs, dim=0)
    i = torch.cat([torch.full((x.shape[0],), idx) for idx, x in enumerate(Xs)]).to(device)
    i_ptr = torch.cat([torch.tensor([0], device=device), i.bincount().cumsum(0)])
    return x, i, i_ptr


def xgower_factor(X):
    a = np.power(X, 2).sum()
    b = X.dot(X.sum(0)).sum()
    return np.sqrt((a - b / X.shape[0]) / (X.shape[0] - 1))


class MILDataset(Dataset):
    def __init__(self, Xs, F, Y):
        self.Xs = Xs
        self.F = F
        self.Y = Y

    def __len__(self):
        return len(self.Xs)

    def __getitem__(self, idx):
        X = self.Xs[idx]
        F = self.F[idx]
        Y = self.Y[idx]
        return X, F, Y


def mil_collate_fn(batch):
    X = [item[0] for item in batch]
    F = torch.stack([item[1] for item in batch])
    Y = torch.stack([item[2] for item in batch])
    return X, F, Y


def normalize_feats(X, norm_factor="std_sqrt"):
    assert norm_factor in ["std", "std_sqrt"]
    train_data = (
        torch.cat(X["train"], dim=0) if isinstance(X["train"], list) else X["train"].reshape(-1, X["train"].shape[2])
    )
    mean = train_data.mean(0, keepdims=True)
    std = train_data.std(0, keepdims=True)
    factor = std * np.sqrt(train_data.shape[1]) if norm_factor == "std_sqrt" else std

    for key in X:
        X[key] = [(x - mean) / factor for x in X[key]] if isinstance(X[key], list) else (X[key] - mean) / factor

    return X


def load_data(dataset="simulation", norm_factor="std_sqrt", **kwargs):
    if dataset == "simulation":
        X, F, Y, u, w = load_simulation(**kwargs)
        X = normalize_feats(X, norm_factor)
    else:
        raise ValueError(f"Unknown dataset: {dataset}")
    return X, F, Y, u, w



import numpy as np
import torch
from torch.distributions import Binomial, Categorical, LowRankMultivariateNormal
from torch.distributions.kl import kl_divergence
from torch.utils.data import DataLoader
from torch_scatter import scatter_softmax, segment_add_csr
from tqdm.auto import trange

from mixmil.data import MILDataset, mil_collate_fn, setup_scatter
from mixmil.posterior import GaussianVariationalPosterior
from mixmil.utils import get_init_params


class MixMIL(torch.nn.Module):
    """Attention-based Multi-instance Mixed Models

    https://arxiv.org/abs/2311.02455

    """

    def __init__(self, Q, K, P=1, likelihood="binomial", n_trials=2, mean_field=False, init_params=None):
        r"""Initialize the MixMil class.

        Parameters:
        - Q (int): The dimension of the latent space.
        - K (int): The number of fixed effects.
        - P (int): The number of outputs.
        - likelihood (str, optional): The likelihood to use. Either "binomial" or "categorical". Default is "binomial".
        - n_trials (int, optional): Number of trials for binomial likelihood. Not used for categorical. Default is 2.
        - mean_field (bool, optional): Toggle mean field approximation for the posterior. Default is False.
        - init_params (tuple, optional): Tuple of (mean, var, var_z, alpha) to initialize the model. Default is None.
            mean (torch.Tensor): The mean of the posterior. Shape: (Q, P). d
            var (torch.Tensor): The variance of the posterior. Shape: (Q, P).
            var_z (torch.Tensor): The $\sigma_{\beta}^2$ hparam of the prior.
                Shape: (1, P) with separate and (1, 1) with shared priors .
            alpha (torch.Tensor): The fixed effect parameters. Shape: (K, P).
        """
        super().__init__()
        self.Q = Q

        alpha = torch.zeros((K, P))
        log_sigma_u = torch.full((1, P), 0.5 * np.log(0.5))
        log_sigma_z = torch.full((1, P), 0.5 * np.log(0.5))

        if init_params is not None:
            *_, var_z, alpha = init_params
            log_sigma_z = 0.5 * torch.log(var_z)

        self.alpha = torch.nn.Parameter(alpha)
        self.log_sigma_u = torch.nn.Parameter(log_sigma_u)
        self.log_sigma_z = torch.nn.Parameter(log_sigma_z)

        self.posterior = GaussianVariationalPosterior(2 * Q, P, mean_field, init_params)

        self.likelihood_name = likelihood
        self.n_trials = n_trials if likelihood == "binomial" else None
        self.is_trained = False

    def init_with_mean_model(Xs, F, Y, likelihood="binomial", n_trials=None, mean_field=False):
        assert (likelihood == "binomial" and n_trials is not None and 0 < n_trials <= 2) or (
            likelihood == "categorical" and n_trials is None
        ), f"n_trials must be 1 or 2 to initialize with binomial mean model, got {n_trials=} and {likelihood=}"
        init_params = get_init_params(Xs, F, Y, likelihood, n_trials)
        Q, K, P = Xs[0].shape[1], F.shape[1], init_params[0].shape[1]
        return MixMIL(Q, K, P, likelihood, n_trials, mean_field, init_params)

    @property
    def prior_distribution(self):
        device = self.log_sigma_u.device
        scale_u = self.log_sigma_u.T * torch.ones([1, self.Q], device=device)
        scale_z = self.log_sigma_z.T * torch.ones([1, self.Q], device=device)
        cov_logdiag = 2 * torch.cat([scale_u, scale_z], 1)
        cov_factor = torch.zeros_like(cov_logdiag)[:, :, None]
        mu = torch.zeros_like(cov_logdiag)
        return LowRankMultivariateNormal(mu, cov_factor, torch.exp(cov_logdiag))

    @property
    def posterior_distribution(self):
        return self.posterior.distribution

    @property
    def qu_mu(self):
        return self.posterior.q_mu[: self.Q]

    @property
    def qz_mu(self):
        return self.posterior.q_mu[self.Q :]

    def likelihood(self, logits, y):
        if self.likelihood_name == "binomial":
            return Binomial(total_count=self.n_trials, logits=logits).log_prob(y[:, :, None]).sum(1).mean()
        elif self.likelihood_name == "categorical":
            logits = logits.permute(0, 2, 1)
            if logits.shape[-1] == 1:
                logits = torch.cat([-logits, logits], 2)
            return Categorical(logits=logits).log_prob(y).mean()

    def loss(self, u, f, y, kld_w=1.0, return_dict=False):
        logits = f.mm(self.alpha)[:, :, None] + u

        ll = self.likelihood(logits, y)
        kld = kl_divergence(self.posterior_distribution, self.prior_distribution)
        kld_term = kld_w * kld.sum() / y.shape[0]
        loss = -ll + kld_term
        if return_dict:
            return loss, dict(loss=loss.item(), ll=ll.item(), kld=kld_term.item())
        return loss

    def get_betas(self, n_samples=None, predict=False):
        assert not (n_samples and predict)
        if n_samples:
            beta = self.posterior.sample(n_samples)
            beta_u = beta[: self.Q, :, :]
            beta_z = beta[self.Q :, :, :]
        else:
            beta_u = self.qu_mu[:, :, None]
            beta_z = self.qz_mu[:, :, None]
        return beta_u, beta_z

    def forward(self, Xs, n_samples=8, scaling=None, predict=False):
        beta_u, beta_z = self.get_betas(n_samples, predict)
        b = torch.sqrt((beta_z**2).mean(0, keepdim=True))
        eta = beta_z / b

        if torch.is_tensor(Xs):
            u = self._calc_bag_emb_effect_tensor(beta_u, eta, Xs)
        else:
            u = self._calc_bag_emb_effect_scatter(beta_u, eta, Xs)

        mean, std = (u.mean(0), u.std(0)) if scaling is None else scaling
        if std.isnan().any():
            std = 1
        u = b * (u - mean) / std
        return u

    def _calc_bag_emb_effect_tensor(self, beta_u, eta, Xs):
        _w = torch.einsum("niq,qps->nips", Xs, beta_u)
        w = torch.softmax(_w, dim=1)
        t = torch.einsum("niq,qps->nips", Xs, eta)
        u = torch.einsum("nips,nips->nps", w, t)
        return u

    def _calc_bag_emb_effect_scatter(self, beta_u, eta, Xs):
        x, i, i_ptr = setup_scatter(Xs)

        _w = torch.einsum("iq,qps->ips", x, beta_u)
        w = scatter_softmax(_w, i, dim=0)
        t = torch.einsum("iq,qps->ips", x, eta)
        u = segment_add_csr(w * t, i_ptr)
        return u

    def train(self, X, F, Y, n_epochs=2_000, batch_size=64, lr=1e-3, verbose=True):
        train_loader = DataLoader(
            MILDataset(X, F, Y),
            shuffle=True,
            batch_size=batch_size,
            collate_fn=None if torch.is_tensor(X) else mil_collate_fn,
        )
        optim = torch.optim.Adam(lr=lr, params=self.parameters())

        history = []
        for epoch in trange(1, n_epochs + 1, desc="Epoch", disable=not verbose):
            for step, (xs, f, y) in enumerate(train_loader):
                u = self(xs)
                loss, ldict = self.loss(u, f, y, kld_w=len(xs) / len(Y), return_dict=True)
                ldict["epoch"], ldict["step"] = epoch, step
                history.append(ldict)
                optim.zero_grad()
                loss.backward()
                optim.step()

        self.is_trained = True
        return history

    @torch.inference_mode()
    def predict(self, Xs, scaling=None):
        return self(Xs, n_samples=None, predict=True, scaling=scaling).squeeze(2)

    @torch.inference_mode()
    def get_weights(self, Xs, ravel=False):
        """Get instance weights after and before softmax"""
        beta_u, _ = self.get_betas(predict=True)
        beta_u = beta_u.squeeze(2)  # not taking mcmc samples
        if torch.is_tensor(Xs):
            _w = torch.einsum("niq,qp->nip", Xs, beta_u)
            w = torch.softmax(_w, dim=1)

        else:
            x, i, _ = setup_scatter(Xs)
            _w = torch.einsum("iq,qp->ip", x, beta_u)
            w = scatter_softmax(_w, i, dim=0)

        if ravel:
            w, _w = w.ravel(), _w.ravel()
        elif not torch.is_tensor(Xs):
            _w = [_w[i == idx] for idx in range(len(Xs))]
            w = [w[i == idx] for idx in range(len(Xs))]
        return w, _w

    def extra_repr(self):
        string = f"Q={self.Q}, K={self.alpha.shape[0]}, P={self.alpha.shape[1]}, likelihood={self.likelihood_name}"
        if self.likelihood_name == "binomial":
            string += f", n_trials={self.n_trials}"
        string += f", device={self.alpha.device}, trained={self.is_trained}"
        string += f"\n(alpha): Parameter(shape={tuple(self.alpha.shape)})\n"
        string += f"(log_sigma_u): Parameter(shape={tuple(self.log_sigma_u.shape)})\n"
        string += f"(log_sigma_z): Parameter(shape={tuple(self.log_sigma_z.shape)})"
        return string



import numpy as np
import scipy.stats as st
from sklearn.metrics import roc_auc_score

from mixmil import MixMIL
from mixmil.data import load_data


def calc_metrics(model, X, u, w):
    u_pred = model.predict(X["test"]).cpu().numpy().ravel()
    w_pred = model.get_weights(X["test"])[0].cpu().numpy().ravel()
    rho_bag = st.spearmanr(u_pred, u["test"]).correlation  # bag level correlation
    is_top_instance = (w["test"] > np.quantile(w["test"], 0.90)).long().ravel()
    auc_instance = roc_auc_score(is_top_instance, w_pred)  # instance-retrieval AUC
    return rho_bag, auc_instance


def test_simulation():
    X, F, Y, u, w = load_data(P=1, seed=0)
    model = MixMIL.init_with_mean_model(X["train"], F["train"], Y["train"], likelihood="binomial", n_trials=2)

    start_rho_bag, start_auc_instance = calc_metrics(model, X, u, w)
    model.train(X["train"], F["train"], Y["train"], n_epochs=40)
    end_rho_bag, end_auc_instance = calc_metrics(model, X, u, w)

    # assert that both metrics improved by at least 10%
    assert end_rho_bag > start_rho_bag * 1.1
    assert end_auc_instance > start_auc_instance * 1.1
